
import
  preserves, std/sets, std/tables

type
  Eval* {.preservesRecord: "eval".} = object
    `expr`*: string
    `path`*: string
    `result`*: Value

  AttrSet* = Table[Symbol, Value]
  Realise* {.preservesRecord: "realise".} = object
    `drv`*: string
    `outputs`*: StringSeq

  LegacyPathAttrs* {.preservesDictionary.} = object
    `ca`*: string
    `deriver`*: string
    `narHash`*: string
    `narSize`*: BiggestInt
    `references`*: StringSeq
    `registrationTime`*: BiggestInt
    `sigs`*: StringSet
    `ultimate`*: bool

  Missing* {.preservesRecord: "missing".} = object
    `targets`*: StringSeq
    `willBuild`*: StringSet
    `willSubstitute`*: StringSet
    `unknown`*: StringSet
    `downloadSize`*: BiggestInt
    `narSize`*: BiggestInt

  Narinfo* {.preservesRecord: "narinfo".} = object
    `path`*: string
    `info`*: AttrSet

  FieldKind* {.pure.} = enum
    `int`, `string`
  `Field`* {.preservesOr.} = object
    case orKind*: FieldKind
    of FieldKind.`int`:
        `int`*: BiggestInt

    of FieldKind.`string`:
        `string`*: string

  
  StringSet* = HashSet[string]
  AddToStoreAttrs* {.preservesDictionary.} = object
    `# prefer a set

  deriver`*: string
    `ca`*: string
    `ca-method`*: Symbol
    `eris`*: seq[byte]
    `name`*: string
    `narHash`*: string
    `narSize`*: BiggestInt
    `references`*: StringSeq
    `registrationTime`*: BiggestInt
    `sigs`*: StringSet
    `ultimate`*: bool

  AddToStoreClientAttrs* {.preservesDictionary.} = object
    `ca-method`*: Symbol
    `eris`*: seq[byte]
    `name`*: string
    `references`*: StringSeq

  PathInfo* {.preservesRecord: "path".} = object
    `path`*: string
    `attrs`*: AttrSet

  Build* {.preservesRecord: "nix-build".} = object
    `input`*: string
    `output`*: Value

  Fields* = seq[Field]
  ActionStart* {.preservesRecord: "start".} = object
    `id`*: BiggestInt
    `level`*: BiggestInt
    `type`*: BiggestInt
    `text`*: string
    `fields`*: Fields
    `parent`*: BiggestInt

  Instantiate* {.preservesRecord: "instantiate".} = object
    `expr`*: string
    `options`*: AttrSet
    `result`*: Value

  StringSeq* = seq[string]
  ActionStop* {.preservesRecord: "stop".} = object
    `id`*: BiggestInt

  ActionResult* {.preservesRecord: "result".} = object
    `id`*: BiggestInt
    `type`*: BiggestInt
    `fields`*: Fields

proc `$`*(x: Eval | AttrSet | Realise | LegacyPathAttrs | Missing | Narinfo |
    Field |
    StringSet |
    AddToStoreAttrs |
    AddToStoreClientAttrs |
    PathInfo |
    Build |
    Fields |
    ActionStart |
    Instantiate |
    StringSeq |
    ActionStop |
    ActionResult): string =
  `$`(toPreserves(x))

proc encode*(x: Eval | AttrSet | Realise | LegacyPathAttrs | Missing | Narinfo |
    Field |
    StringSet |
    AddToStoreAttrs |
    AddToStoreClientAttrs |
    PathInfo |
    Build |
    Fields |
    ActionStart |
    Instantiate |
    StringSeq |
    ActionStop |
    ActionResult): seq[byte] =
  encode(toPreserves(x))
